@@ -1,21 +1,38 @@
 /**********************************************************************
  *
  * PostGIS - Spatial Types for PostgreSQL
+ * http://postgis.net
+ *
+ * PostGIS is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation, either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * PostGIS is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with PostGIS.  If not, see <http://www.gnu.org/licenses/>.
+ *
+ **********************************************************************
  *
  * Copyright (C) 2009 Paul Ramsey <pramsey@cleverelephant.ca>
  *
- * This is free software; you can redistribute and/or modify it under
- * the terms of the GNU General Public Licence. See the COPYING file.
- *
  **********************************************************************/
 
-#include "liblwgeom_internal.h"
+
+#include "../postgis_config.h"
+/*#define POSTGIS_DEBUG_LEVEL 4*/
+#include "liblwgeom_internal.h" /* NOTE: includes lwgeom_log.h */
 #include "lwgeom_log.h"
+#include <math.h>
 
 /**
 * Used for passing the parse state between the parsing functions.
 */
-typedef struct 
+typedef struct
 {
 	const uint8_t *wkb; /* Points to start of WKB */
 	size_t wkb_size; /* Expected size of WKB */
@@ -102,18 +119,18 @@ uint8_t* bytes_from_hexbytes(const char *hexbuf, size_t hexsize)
 
 
 /**
-* Check that we are not about to read off the end of the WKB 
+* Check that we are not about to read off the end of the WKB
 * array.
 */
 static inline void wkb_parse_state_check(wkb_parse_state *s, size_t next)
 {
 	if( (s->pos + next) > (s->wkb + s->wkb_size) )
 		lwerror("WKB structure does not match expected size!");
-} 
+}
 
 /**
 * Take in an unknown kind of wkb type number and ensure it comes out
-* as an extended WKB type number (with Z/M/SRID flags masked onto the 
+* as an extended WKB type number (with Z/M/SRID flags masked onto the
 * high bits).
 */
 static void lwtype_from_wkb_state(wkb_parse_state *s, uint32_t wkb_type)
@@ -157,10 +174,10 @@ static void lwtype_from_wkb_state(wkb_parse_state *s, uint32_t wkb_type)
 
 	switch (wkb_simple_type)
 	{
-		case WKB_POINT_TYPE: 
+		case WKB_POINT_TYPE:
 			s->lwtype = POINTTYPE;
 			break;
-		case WKB_LINESTRING_TYPE: 
+		case WKB_LINESTRING_TYPE:
 			s->lwtype = LINETYPE;
 			break;
 		case WKB_POLYGON_TYPE:
@@ -175,7 +192,7 @@ static void lwtype_from_wkb_state(wkb_parse_state *s, uint32_t wkb_type)
 		case WKB_MULTIPOLYGON_TYPE:
 			s->lwtype = MULTIPOLYGONTYPE;
 			break;
-		case WKB_GEOMETRYCOLLECTION_TYPE: 
+		case WKB_GEOMETRYCOLLECTION_TYPE:
 			s->lwtype = COLLECTIONTYPE;
 			break;
 		case WKB_CIRCULARSTRING_TYPE:
@@ -190,7 +207,7 @@ static void lwtype_from_wkb_state(wkb_parse_state *s, uint32_t wkb_type)
 		case WKB_MULTICURVE_TYPE:
 			s->lwtype = MULTICURVETYPE;
 			break;
-		case WKB_MULTISURFACE_TYPE: 
+		case WKB_MULTISURFACE_TYPE:
 			s->lwtype = MULTISURFACETYPE;
 			break;
 		case WKB_POLYHEDRALSURFACE_TYPE:
@@ -318,6 +335,9 @@ static POINTARRAY* ptarray_from_wkb_state(wkb_parse_state *s)
 
 	/* Calculate the size of this point array. */
 	npoints = integer_from_wkb_state(s);
+
+	LWDEBUGF(4,"Pointarray has %d points", npoints);
+
 	if( s->has_z ) ndims++;
 	if( s->has_m ) ndims++;
 	pa_size = npoints * ndims * WKB_DOUBLE_SIZE;
@@ -353,10 +373,10 @@ static POINTARRAY* ptarray_from_wkb_state(wkb_parse_state *s)
 
 /**
 * POINT
-* Read a WKB point, starting just after the endian byte, 
+* Read a WKB point, starting just after the endian byte,
 * type number and optional srid number.
 * Advance the parse state forward appropriately.
-* WKB point has just a set of doubles, with the quantity depending on the 
+* WKB point has just a set of doubles, with the quantity depending on the
 * dimension of the point, so this looks like a special case of the above
 * with only one point.
 */
@@ -366,6 +386,7 @@ static LWPOINT* lwpoint_from_wkb_state(wkb_parse_state *s)
 	POINTARRAY *pa = NULL;
 	size_t pa_size;
 	uint32_t ndims = 2;
+	const POINT2D *pt;
 
 	/* Count the dimensions. */
 	if( s->has_z ) ndims++;
@@ -394,14 +415,24 @@ static LWPOINT* lwpoint_from_wkb_state(wkb_parse_state *s)
 		}
 	}
 	
-	return lwpoint_construct(s->srid, NULL, pa);
+	/* Check for POINT(NaN NaN) ==> POINT EMPTY */
+	pt = getPoint2d_cp(pa, 0);
+	if ( isnan(pt->x) && isnan(pt->y) )
+	{
+		ptarray_free(pa);
+		return lwpoint_construct_empty(s->srid, s->has_z, s->has_m);
+	}
+	else
+	{
+		return lwpoint_construct(s->srid, NULL, pa);
+	}
 }
 
 /**
 * LINESTRING
-* Read a WKB linestring, starting just after the endian byte, 
-* type number and optional srid number. Advance the parse state 
-* forward appropriately. 
+* Read a WKB linestring, starting just after the endian byte,
+* type number and optional srid number. Advance the parse state
+* forward appropriately.
 * There is only one pointarray in a linestring. Optionally
 * check for minimal following of rules (two point minimum).
 */
@@ -423,9 +454,9 @@ static LWLINE* lwline_from_wkb_state(wkb_parse_state *s)
 
 /**
 * CIRCULARSTRING
-* Read a WKB circularstring, starting just after the endian byte, 
-* type number and optional srid number. Advance the parse state 
-* forward appropriately. 
+* Read a WKB circularstring, starting just after the endian byte,
+* type number and optional srid number. Advance the parse state
+* forward appropriately.
 * There is only one pointarray in a linestring. Optionally
 * check for minimal following of rules (three point minimum,
 * odd number of points).
@@ -454,9 +485,9 @@ static LWCIRCSTRING* lwcircstring_from_wkb_state(wkb_parse_state *s)
 
 /**
 * POLYGON
-* Read a WKB polygon, starting just after the endian byte, 
-* type number and optional srid number. Advance the parse state 
-* forward appropriately. 
+* Read a WKB polygon, starting just after the endian byte,
+* type number and optional srid number. Advance the parse state
+* forward appropriately.
 * First read the number of rings, then read each ring
 * (which are structured as point arrays)
 */
@@ -465,6 +496,8 @@ static LWPOLY* lwpoly_from_wkb_state(wkb_parse_state *s)
 	uint32_t nrings = integer_from_wkb_state(s);
 	int i = 0;
 	LWPOLY *poly = lwpoly_construct_empty(s->srid, s->has_z, s->has_m);
+
+	LWDEBUGF(4,"Polygon has %d rings", nrings);
 	
 	/* Empty polygon? */
 	if( nrings == 0 )
@@ -505,9 +538,9 @@ static LWPOLY* lwpoly_from_wkb_state(wkb_parse_state *s)
 
 /**
 * TRIANGLE
-* Read a WKB triangle, starting just after the endian byte, 
-* type number and optional srid number. Advance the parse state 
-* forward appropriately. 
+* Read a WKB triangle, starting just after the endian byte,
+* type number and optional srid number. Advance the parse state
+* forward appropriately.
 * Triangles are encoded like polygons in WKB, but more like linestrings
 * as lwgeometries.
 */
@@ -553,9 +586,9 @@ static LWTRIANGLE* lwtriangle_from_wkb_state(wkb_parse_state *s)
 	}
 
 	/* Empty TRIANGLE starts w/ empty POINTARRAY, free it first */
-	if ( tri->points) 
+	if (tri->points)
 		ptarray_free(tri->points);
-
+	
 	tri->points = pa;	
 	return tri;
 }
@@ -590,7 +623,7 @@ static LWCURVEPOLY* lwcurvepoly_from_wkb_state(wkb_parse_state *s)
 
 /**
 * COLLECTION, MULTIPOINTTYPE, MULTILINETYPE, MULTIPOLYGONTYPE, COMPOUNDTYPE,
-* MULTICURVETYPE, MULTISURFACETYPE, 
+* MULTICURVETYPE, MULTISURFACETYPE,
 * TINTYPE
 */
 static LWCOLLECTION* lwcollection_from_wkb_state(wkb_parse_state *s)
@@ -599,6 +632,8 @@ static LWCOLLECTION* lwcollection_from_wkb_state(wkb_parse_state *s)
 	LWCOLLECTION *col = lwcollection_construct_empty(s->lwtype, s->srid, s->has_z, s->has_m);
 	LWGEOM *geom = NULL;
 	int i;
+
+	LWDEBUGF(4,"Collection has %d components", ngeoms);
 	
 	/* Empty collection? */
 	if ( ngeoms == 0 )
@@ -719,10 +754,10 @@ LWGEOM* lwgeom_from_wkb_state(wkb_parse_state *s)
 /**
 * WKB inputs *must* have a declared size, to prevent malformed WKB from reading
 * off the end of the memory segment (this stops a malevolent user from declaring
-* a one-ring polygon to have 10 rings, causing the WKB reader to walk off the 
+* a one-ring polygon to have 10 rings, causing the WKB reader to walk off the
 * end of the memory).
 *
-* Check is a bitmask of: LW_PARSER_CHECK_MINPOINTS, LW_PARSER_CHECK_ODD, 
+* Check is a bitmask of: LW_PARSER_CHECK_MINPOINTS, LW_PARSER_CHECK_ODD,
 * LW_PARSER_CHECK_CLOSURE, LW_PARSER_CHECK_NONE, LW_PARSER_CHECK_ALL
 */
 LWGEOM* lwgeom_from_wkb(const uint8_t *wkb, const size_t wkb_size, const char check)
@@ -742,7 +777,7 @@ LWGEOM* lwgeom_from_wkb(const uint8_t *wkb, const size_t wkb_size, const char ch
 	s.pos = wkb;
 	
 	/* Hand the check catch-all values */
-	if ( check & LW_PARSER_CHECK_NONE ) 
+	if ( check & LW_PARSER_CHECK_NONE )
 		s.check = 0;
 	else
 		s.check = check;
